/* ceh.c - critical handler shell */

#include<stdio.h>
#include<dos.h>
#include<stdarg.h>
#include<stdlib.h>
#include<conio.h>
#include<ctype.h>
#include<string.h>
#if !defined(TRUE)
#define TRUE 1
#endif

#if !defined(FALSE)
#define FALSE 0
#endif

void bios_printf(char *fmt,...);
void far critical_error_handler(unsigned deverr,
    unsigned errcode, unsigned far *devhdr);

void main(void)
    {
    FILE *fp;

    printf("Installing Critical Error Handler\n");
    _harderr(critical_error_handler);
    printf("Opening file for reading on "
           "drive A...\n");
    if(!(fp = fopen("a:newfile.txt","r")))
        printf(
            "Unable to open file for reading\n");
    else
        fclose(fp);

    printf(
        "Opening file for writing on "
        "drive A...\n");
    if(!(fp = fopen("a:newfile.txt","w")))
        printf(
            "Unable to open file for writing\n");
    printf("Exiting...\n\r");
    exit(0);
    }

/* critical error handler messages */
char *ceh_errmsg[] =
    {
    "Attempt to write to write-protected disk",
    "Unknown unit",
    "Drive not ready",
    "Unknown command",
    "CRC error in data",
    "Bad drive-request structure length",
    "Seek error",
    "Unknown media type",
    "Sector not found",
    "Printer out of paper",
    "Write fault",
    "Read fault",
    "General failure"
    };

char *ceh_location[] = 
    {
    "MS-DOS",
    "FAT",
    "Directory",
    "Data area"
    };

char *ceh_device[] =
    {
    "clock device",
    "null device",
    "standard output",
    "standard input"
    };

/* critical error handler function */
void far critical_error_handler(unsigned deverr,
    unsigned errcode, unsigned far *devhdr)
    {
    unsigned i,j,codelocation,drive,loc,c,
            prev = FALSE, fail_allowed = FALSE,
            attribute_word;
    char *action,*comma_space = ", ",
            *nullstr = "",buf[80];
    char far *hdr;

    if(deverr & 0x8000) /* if not a disk error  */
        {
        hdr = (char far *)devhdr;
        hdr += 4;
        attribute_word = *((unsigned far *)hdr);
        if(attribute_word & 0x8000)
            sprintf(buf,
                "Bad File Allocation Table\n\r");
        else
            {
            buf[0] = '\0';
            for(i = 1,j = 0; i < 0x0010;
                i <<= 1,j++)
                if(attribute_word & i)
                    sprintf(&buf[strlen(buf)],
                        "Error on Current %s\n\r",
                        ceh_device[j]);
            }
        }
    else    /* disk error   */
        {
        action  = ((deverr & 0x0100)
            ? "writing" : "reading");
        loc = ((deverr & 0x0600) >> 9);
        drive = (deverr & 0x00ff)+'A';
        sprintf(buf,
            "Error in %s, %s drive %c, %s\n\r",
            ceh_location[loc],action,drive,
            ceh_errmsg[(errcode & 0x00ff)]);
        }

    /* stay in loop until user selects option */
    while(TRUE)
        {
        cprintf(buf);
        if(deverr & 0x0800)
            {
            prev = TRUE;
            cprintf("Abort");
            if(_osmajor > 2)
                fail_allowed = TRUE;
            }
        if(deverr & 0x1000)
            {
            cprintf("%sRetry",
                (prev ? comma_space : nullstr));
            prev = TRUE;
            }
        if(deverr & 0x2000)
            {
            cprintf("%sIgnore",
                (prev ? comma_space : nullstr));
            prev = TRUE;
            }
        if(fail_allowed)
            cprintf("%sFail",
                (prev ? comma_space : nullstr));
        cprintf("? ");
        c = getche();
        cprintf("\n\r");
        switch(toupper(c))
            {
            case 'A':
                _hardresume(_HARDERR_ABORT);
                break;
            case 'R':
                _hardresume(_HARDERR_RETRY);
                break;
            case 'I':
                _hardresume(_HARDERR_IGNORE);
                break;
            case 'F':
                if(fail_allowed)
                    _hardresume(_HARDERR_FAIL);
            }
        }
    }
